////////////////////////////////////////////////////////////////////////////
// Fragment shader to compute per-vertex elements of differential geometry 
// of 3rd order.
//
// Copyright (C) 2007 Harlen Costa Batagelo. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the 
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Global variables
////////////////////////////////////////////////////////////////////////////
int g_iAdjVtMapSize;  // Size of vertex adjacency map
int g_iVtMapSize;     // Size of the vertex attribute map

uniform sampler2D s0; // Vertex position
uniform sampler2D s2; // Adjacency address
uniform sampler2D s3; // Vertex adjacency
uniform sampler2D s4; // Principal direction and curvature k1
uniform sampler2D s5; // Principal direction and curvature k2

const float M_2_PI = 0.636619772367581343076;
const float M_PI_2 = 1.57079632679489661923;
const float M_PI = 3.14159265;

vec3 hsv(vec3 hsv)
{
	// From FvD
	hsv.x = mod(hsv.x, 2.0 * M_PI);
	if (hsv.x < 0.0)
		hsv.x += 2.0 * M_PI;
	hsv.x /= M_PI / 3.0;
	int i = int(floor(hsv.x));
	float f = hsv.x - float(i);
	float p = hsv.z * (1.0 - hsv.y);
	float q = hsv.z * (1.0 - (hsv.y*f));
	float t = hsv.z * (1.0 - (hsv.y*(1.0-f)));
	
	if( i == 0 ) return vec3(hsv.z, t, p);
	else if( i == 1 ) return vec3(q, hsv.z, p);
	else if( i == 2 ) return vec3(p, hsv.z, t);
	else if( i == 3 ) return vec3(p, q, hsv.z);
	else if( i == 4 ) return vec3(t, p, hsv.z);
	else return vec3(hsv.z, p, q);
}

////////////////////////////////////////////////////////////////////////////
// Perform LDL^T decomposition of a 4x4 symmetric positive definite matrix
// and solves Ax=B.
////////////////////////////////////////////////////////////////////////////
vec4 ldltsv4( mat4 w, vec4 m )
{
	// Decompose matrix. Overwrites lower triangle of matrix.
	vec4 d;
	vec3 v;
	int i;
	for( i = 0; i < 4; i++ ) 
	{
		for( int k = 0; k < i; k++ ) v[k] = w[i][k] * d[k];
		for( int j = i; j < 4; j++ ) 
		{
			float sum = w[i][j];
			for( int k = 0; k < i; k++ ) 
				sum -= v[k] * w[j][k];
			if( i == j ) 
				d[i] = 1.0 / sum;
			else
				w[j][i] = sum;
		}
	}
	
	// Solve Ax=B after LDL^T decomposition
	for( i = 0; i < 4; i++ )
	{
		float sum = m[i];
		for( int k = 0; k < i; k++ ) 
			sum -= w[i][k] * m[k];
		m[i] = sum * d[i];
	}
	for( i = 3; i >= 0; i-- )
	{
		float sum = 0.0;
		for( int k = i + 1; k < 4; k++ ) 
			sum += w[k][i] * m[k];
		m[i] -= sum * d[i];
	}
	
	return m;
}

////////////////////////////////////////////////////////////////////////////
// Convert 1D index (i) to 2D UV index.
// i = 1D index
// s = size of a screen line, in pixels
////////////////////////////////////////////////////////////////////////////
vec2 I1DTo2D(float i, float s) { return vec2(mod(i,s),int(i/s)); }

////////////////////////////////////////////////////////////////////////////
// Pixel shader for computing 3rd fundamental forms.
////////////////////////////////////////////////////////////////////////////
void main( )
{
	// Sample adjacency addresses
	// .x) Address to the 1st adjacent face in the face adjacency texture
	// .y) Number of adjacent faces
	// .z) Address to the 1st adjacent vertex in the vertex adjacency 
	//     texture
	// .w) Number of adjacent vertices
	vec4 fAdj = texture2D( s2, gl_TexCoord[0].st );
	if( fAdj.z == -1.0 ) discard;
	
	vec3 vPos   = texture2D( s0, gl_TexCoord[0].st ).xyz;	
	vec4 vpk1   = texture2D( s4, gl_TexCoord[0].st ); 
	vec4 vpk2   = texture2D( s5, gl_TexCoord[0].st ); 	

	vec3 vPDir1 = vpk1.xyz, vPDir2 = vpk2.xyz;
	vec2 vCurv  = vec2( vpk1.w, vpk2.w );
	vec2 vCurvNei;

	vec4 m = vec4( 0.0 );	
	mat4 w = mat4( 0.0 );
	
	float fCt = 0.0;
	while( fCt < fAdj.w )
    {
		// Get neighbor index from the vertex adjacency texture
		vec2  vAdjUVNei  = I1DTo2D( fAdj.z + fCt, float(g_iAdjVtMapSize) );		
		float fIdxNei    = texture2D( s3, vAdjUVNei ).x;
		
		// Get attributes of neighbor vertex		
		vec2 vUVNei      = I1DTo2D( fIdxNei, float(g_iVtMapSize) );		
		vec3 vPosNei     = texture2D( s0, vUVNei ).xyz;
		vCurvNei.x       = texture2D( s4, vUVNei ).w;
		vCurvNei.y       = texture2D( s5, vUVNei ).w;
    
		vec3 fcurv_i = vec3( vCurv.x,    0, vCurv.y    ), 
		     fcurv_n = vec3( vCurvNei.x, 0, vCurvNei.y );
    
		// Variation of curvature along each edge
		vec3 dfcurv = fcurv_n - fcurv_i;
		vec3 e = vPosNei - vPos;

		float u = dot( e, vPDir1 );
		float v = dot( e, vPDir2 );
		float u2 = u*u, v2 = v*v, uv = u*v;
		w[0][0] += u2;
		w[0][1] += uv;
		w[3][3] += v2;
		m[0] += u*dfcurv[0];
		m[1] += v*dfcurv[0] + 2.0*u*dfcurv[1];
		m[2] += 2.0*v*dfcurv[1] + u*dfcurv[2];
		m[3] += v*dfcurv[2];
    	
		fCt++;
    } 
    
	w[1][1] = 2.0 * w[0][0] + w[3][3];
	w[1][2] = 2.0 * w[0][1];
	w[2][2] = w[0][0] + 2.0 * w[3][3];
	w[2][3] = w[0][1];
    
	m = ldltsv4( w, m );
	
	// 3rd fundamental form
	gl_FragColor = m;
	
/*
	float cscale = 1.0;
	float H = 0.25 * (m[0] + m[1] + m[2] + m[3]);
	float K = m[0] * m[1] * m[2] * m[3];
	
	float h = 4.0 / 3.0f * abs(atan((H*H-K)/(H*H*sign(H))));
	float sc = M_2_PI * atan((2.0*H*H-K)*cscale);
	
	vec3 teste = vec3(h,sc,1.0);
	if( teste.y <= 0.0)
		gl_FragColor = vec4( teste.z, teste.z, teste.z, 1.0 );
	else	
		gl_FragColor = vec4( hsv(teste), 1.0 );
		
	float teste = int(min(max(256.0 / M_PI * (atan(H*cscale) + M_PI_2),
			    0.0), 255.99));
	gl_FragColor = vec4(teste / 255, teste / 255, teste / 255,1.0);
*/
}