////////////////////////////////////////////////////////////////////////////////////////
// File: GPUIntLib.cpp
//
// Sample application for demonstrating the GPU interaction library.
//
// See Copyright Notice at the end of the file.
////////////////////////////////////////////////////////////////////////////////////////

#include "CIntManager.h"

#include <GL/glut.h>
#include <stdlib.h>

// ROI size
#define ROIW 1
#define ROIH 1

CIntManager *g_pIntManager = NULL;
float *g_pROI[1]; // Region of interest

float g_fNear =  1.0;
float g_fFar  = 20.0;

// Callback function used during the rendering pass for 
// updating the vertex attributes of the geometry
void UpdateCallBack( CIntObj *pObj, void *pUserData )
{
  glPushMatrix( );
  glBegin( GL_TRIANGLES );
    glColor4f( 0.0, 0.0, 0.0, 0.0 ); // Vertex ID
    glTexCoord4i( 1, 1, 1, 1 ); // Model ID
    glVertex3f( -2.0, -2.0, 0.0 );

	glColor4f( 1.0, 1.0, 1.0, 1.0 ); // Vertex ID
	glTexCoord4i( 1, 1, 1, 1 ); // Model ID
    glVertex3f( 2.0, -1.0, 0.0 );

	glColor4f( 2.0, 2.0, 2.0, 2.0 ); // Vertex ID
	glTexCoord4i( 1, 1, 1, 1 ); // Model ID
    glVertex3f( 0.0, 2.0, 0.0 );
  glEnd( );
  glPopMatrix( );
}

// Callback function used for rendering geometry.
void RenderCallBack( CIntObj *pObj, int iRP, void *pUserData )
{
  glPushMatrix( );
  glBegin( GL_TRIANGLES );
    glColor4f( 0.0, 0.0, 0.0, 0.0 ); // Vertex ID
    glTexCoord4i( 1, 1, 1, 1 ); // Model ID
    glVertex3f( -2.0, -2.0, 0.0 );

	glColor4f( 1.0, 1.0, 1.0, 1.0 ); // Vertex ID
	glTexCoord4i( 1, 1, 1, 1 ); // Model ID
    glVertex3f( 2.0, -1.0, 0.0 );

	glColor4f( 2.0, 2.0, 2.0, 2.0 ); // Vertex ID
	glTexCoord4i( 1, 1, 1, 1 ); // Model ID
    glVertex3f( 0.0, 2.0, 0.0 );
  glEnd( );
  glPopMatrix( );
}


void Init( )
{
  // Create instance of architecture of interaction
  g_pIntManager = new CIntManager( );

  // Set desired interaction attributes
  ATTRIBUTE m_stAttribute[] =
  {
    ATTTYPE_USERDEFI, // Model ID
    ATTTYPE_END
  };
  g_pIntManager->SetAttributes( m_stAttribute );

  // Set semantic bindings
  SEMANTICBINDING stBinding[] =
  {
    // Vertex ID is in the vertex color
    { VSSEMANTIC_COLOR0, INTSEMANTIC_VERTEXID  },
    // Model ID is in the first set of texture coordinates
    { VSSEMANTIC_TEXCOORD0, INTSEMANTIC_USERDEFI  },
    SEMANTIC_END
  };
  g_pIntManager->BindSemantics( stBinding );

  // Create instance of interaction object
  CIntObj *pMyObj = NULL;
  g_pIntManager->CreateObject( &pMyObj, NULL, NULL );

  // Set callback functions
  pMyObj->SetUpdateCallback( UpdateCallBack );
  pMyObj->SetRenderCallback( RenderCallBack, NULL );

  // Set shader for changing vertex attributes before the
  // computation of diff. geom. properties (stage 1)
  static char *szVSPreDeform = {
    "VSDEFORM fPreVSDeform( VSDEFORM In )                      \n"
    "{                                                         \n"
    "  VSDEFORM Out = In;                                      \n"
    "  // Functions for changing Out attributes go here        \n"
    "  return Out;                                             \n"
    "}                                                         \n"
  };
  pMyObj->SetPreVertexDeform( szVSPreDeform );

  // Set shader for changing vertex attributes after the
  // computation of diff. geom. properties (stage 2)
  static char *szVSPostDeform = {
    "PSDEFORM fPostVSDeform( PSDEFORM In )                     \n"
    "{                                                         \n"
    "  PSDEFORM Out = In;                                      \n"
    "  // Functions for changing Out attributes go here        \n"
    "  return Out;                                             \n"
    "}                                                         \n"
  };
  pMyObj->SetPostVertexDeform( szVSPostDeform );

  // Set shader for changing fragment attributes
  static char *szPSDeform = {
    "PSDEFORM fPSDeform( PSDEFORM In )                         \n"
    "{                                                         \n"
    "  PSDEFORM Out = In;                                      \n"
    "  // Functions for changing Out attributes go here        \n"
    "  return Out;                                             \n"
    "}                                                         \n"
  };
  pMyObj->SetPixelDeform( szPSDeform );

  // Set vertex and index buffer of the geometry
  float pVtPos[9] = { -2.0, -2.0, 0.0,
                       2.0, -1.0, 0.0,
                       0.0,  2.0, 0.0 };
  int pIdx[3] = { 0, 1, 2 };

  pMyObj->SetVertexBuffer( pVtPos, NULL, NULL, 3 );
  pMyObj->SetIndexBuffer( pIdx, 1 );

  glEnable( GL_DEPTH_TEST );
  glClearColor( 0.0, 0.0, 0.0, 0.0 );

  // Allocate memory for the region of interest
  g_pROI[0] = new float[ROIW*ROIH*4]; // 1 pixel with 1 RGBA attribute

  g_pIntManager->SetROI( 0, 0, ROIW-1, ROIH-1 );
}

void Display( )
{
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  // Render model for visualization
  RenderCallBack( NULL, 0, NULL );

  // Run the architecture processing pipeline and render
  // the model onto the offscreen render targets
  g_pIntManager->Render( );

  // Decode attributes
  g_pIntManager->Decode( g_pROI );

  //g_pIntManager->ShowRT( );

  glutSwapBuffers( );

  if( g_pROI[0][0] )
    printf( "Selected model!\n" );
}

void Reshape( int iW, int iH )
{
  glViewport( 0, 0, (GLsizei) iW, (GLsizei) iH );
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity( );
  gluPerspective( 60.0, (GLfloat) iW/(GLfloat) iH, g_fNear, g_fFar );
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity( );
  gluLookAt( 0.0, 0.0, 5.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0 );

  g_pIntManager->SetZNear( g_fNear );
  g_pIntManager->SetZFar ( g_fFar  );
}

void Mouse( int iX, int iY )
{
  // Region of interest encompasses only the pixel pointed by the cursor
  g_pIntManager->SetROI( iX, iY, iX+ROIW-1, iY+ROIH-1 );
}

int main( int argc, char** argv )
{
  glutInit( &argc, argv );
  glutInitDisplayMode( GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA );
  glutInitWindowSize( 600, 600 );
  glutInitWindowPosition( 0, 0 );
  glutCreateWindow( argv[0] );

  GLenum err = glewInit( );
  if( GLEW_OK != err )
  {
    // Problem: glewInit failed, something is seriously wrong.
    fprintf( stderr, "Error: %s\n", glewGetErrorString( err ) );
  }

  fprintf( stdout, "Status: Using GLEW %s\n", glewGetString( GLEW_VERSION ) );

  Init( );
  glutDisplayFunc( Display );
  glutIdleFunc( Display );
  glutReshapeFunc( Reshape );
  glutPassiveMotionFunc( Mouse );
  glutMainLoop( );

  return 0;
}

////////////////////////////////////////////////////////////////////////////////
/// Copyright (C) 2007 Harlen Costa Batagelo. All rights reserved.
///
/// Permission is hereby granted, free of charge, to any person obtaining
/// a copy of this software and associated documentation files (the 
/// "Software"), to deal in the Software without restriction, including
/// without limitation the rights to use, copy, modify, merge, publish,
/// distribute, sublicense, and/or sell copies of the Software, and to
/// permit persons to whom the Software is furnished to do so, subject to
/// the following conditions:
///
/// The above copyright notice and this permission notice shall be
/// included in all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
/// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
/// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
/// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
/// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
/// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
/// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////